abbonement	abonnement
addresse	adresse
agresiv	aggressiv
aggresiv	aggressiv
agressiv	aggressiv
aldrig	aldri
alldri	aldri
allè	allé
allminnelig	alminnelig
alminnlig	alminnelig
altid	alltid
annerkjenne	anerkjenne
apelere	appellere
apellere	appellere
appelere	appellere
apetitt	appetitt
appetit	appetitt
artikel	artikkel
artikkler	artikler
avlegs	avleggs
biblotek	bibliotek
bilett	billett
blandt	blant
brylupp	bryllup
colloseum	colosseum
desverre	dessverre
editere	redigere
ein	en
eit	et
fasinerende	fascinerende
foresten	forresten
genetiv	genitiv
gjordt	gjort
grafitti	graffiti
hovedsaklig	hovedsakelig
ikkje	ikke
iakta	iaktta
insjiativ	initiativ
inttiativ	initiativ
intiativ	initiativ
inteligent	intelligent
intligent	intelligent
innteligent	intelligent
internett	Internett
Internet	Internett
interresant	interessant
interressant	interessant
interrese	interesse
interresse	interesse
km2	km²
kakkerlakk	kakerlakk
kammerat	kamerat
kammere	kamre
kanate	kantate
kansje	kanskje
karriære	karrière
kariere	karrière
kasett	kassett
komune	kommune
konstantere	konstatere
kontigent	kontingent
konkuranse	konkurranse
konkurere	konkurrere
m.o.h.	moh
Majorstua	Majorstuen
media	medier
medlemsskap	medlemskap
mennsker	mennesker
metrolog	meteorolog
metrologi	meteorologi
Midt-Østen	Midtøsten
Midt-østen	Midtøsten
millenium	millennium
missvisende	misvisende
morro	moro
musikker	musiker
nysjerrig	nysgjerrig
nyskjerrig	nysgjerrig
oksyd	oksid
oksydere	oksidere
orginal	original
paralell	parallell
parallel	parallell
parantes	parentes
prevansjon	prevensjon
risici	risikoer
risiki	risikoer
satelitt	satellitt
simpelten	simpelthen
sinnsyk	sinnssyk
skjelden	sjelden
skriftelig	skriftlig
slalom	slalåm
terasse	terrasse
triologi	trilogi
tydlig	tydelig
tunell	tunnel
tunnell	tunnel
untatt	unntatt
usakelig	usaklig
utdybende	utdypende
underholdene	underholdende
vakum	vakuum
viderekommende	viderekomne
værre	verre
værste	verste
